@@ -2,24 +2,33 @@
  * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
  *
  * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Library General Public
+ * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
+ * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU Library General Public
+ * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 02111-1307, USA.
  */
+
+/*
+ * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
+ * file for a list of people on the GTK+ Team.  See the ChangeLog
+ * files for a list of changes.  These files are distributed with
+ * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
+ */
+
 #include <stdlib.h>
 #include <string.h>
 #include <stdio.h>
 
+#include "gtklabel.h"
 #include "gtkmain.h"
 #include "gtkwidget.h"
 #include "gtkwindow.h"
@@ -29,6 +38,12 @@
 
 
 #define DEFAULT_DELAY 500           /* Default delay in ms */
+#define STICKY_DELAY 0              /* Delay before popping up next tip
+                                     * if we're sticky
+                                     */
+#define STICKY_REVERT_DELAY 1000    /* Delay before sticky tooltips revert
+				     * to normal
+                                     */
 
 static void gtk_tooltips_class_init        (GtkTooltipsClass *klass);
 static void gtk_tooltips_init              (GtkTooltips      *tooltips);
@@ -43,31 +58,33 @@ static void gtk_tooltips_widget_remove     (GtkWidget   *widget,
 static void gtk_tooltips_set_active_widget (GtkTooltips *tooltips,
                                             GtkWidget   *widget);
 static gint gtk_tooltips_timeout           (gpointer     data);
-static void gtk_tooltips_create_window     (GtkTooltips *tooltips);
+
+static gint gtk_tooltips_paint_window      (GtkTooltips *tooltips);
 static void gtk_tooltips_draw_tips         (GtkTooltips *tooltips);
 
-static GtkDataClass *parent_class;
+static GtkObjectClass *parent_class;
 static const gchar  *tooltips_data_key = "_GtkTooltipsData";
 
-guint
+GtkType
 gtk_tooltips_get_type (void)
 {
-  static guint tooltips_type = 0;
+  static GtkType tooltips_type = 0;
 
   if (!tooltips_type)
     {
-      GtkTypeInfo tooltips_info =
+      static const GtkTypeInfo tooltips_info =
       {
 	"GtkTooltips",
 	sizeof (GtkTooltips),
 	sizeof (GtkTooltipsClass),
 	(GtkClassInitFunc) gtk_tooltips_class_init,
 	(GtkObjectInitFunc) gtk_tooltips_init,
-	(GtkArgSetFunc) NULL,
-	(GtkArgGetFunc) NULL,
+	/* reserved_1 */ NULL,
+	/* reserved_2 */ NULL,
+        (GtkClassInitFunc) NULL,
       };
 
-      tooltips_type = gtk_type_unique (gtk_data_get_type (), &tooltips_info);
+      tooltips_type = gtk_type_unique (GTK_TYPE_OBJECT, &tooltips_info);
     }
 
   return tooltips_type;
@@ -79,7 +96,7 @@ gtk_tooltips_class_init (GtkTooltipsClass *class)
   GtkObjectClass *object_class;
 
   object_class = (GtkObjectClass*) class;
-  parent_class = gtk_type_class (gtk_data_get_type ());
+  parent_class = gtk_type_class (GTK_TYPE_OBJECT);
 
   object_class->destroy = gtk_tooltips_destroy;
 }
@@ -90,26 +107,19 @@ gtk_tooltips_init (GtkTooltips *tooltips)
   tooltips->tip_window = NULL;
   tooltips->active_tips_data = NULL;
   tooltips->tips_data_list = NULL;
-  tooltips->gc = NULL;
-  tooltips->foreground = NULL;
-  tooltips->background = NULL;
   
   tooltips->delay = DEFAULT_DELAY;
   tooltips->enabled = TRUE;
   tooltips->timer_tag = 0;
+  tooltips->use_sticky_delay = FALSE;
+  tooltips->last_popdown.tv_sec = -1;
+  tooltips->last_popdown.tv_usec = -1;
 }
 
 GtkTooltips *
 gtk_tooltips_new (void)
 {
-  return gtk_type_new (gtk_tooltips_get_type ());
-}
-
-void
-gtk_tooltips_free_string (gpointer data, gpointer user_data)
-{
-  if (data)
-    g_free (data);
+  return gtk_type_new (GTK_TYPE_TOOLTIPS);
 }
 
 static void
@@ -117,9 +127,6 @@ gtk_tooltips_destroy_data (GtkTooltipsData *tooltipsdata)
 {
   g_free (tooltipsdata->tip_text);
   g_free (tooltipsdata->tip_private);
-  g_list_foreach (tooltipsdata->row, gtk_tooltips_free_string, 0);
-  if (tooltipsdata->row)
-    g_list_free (tooltipsdata->row);
   gtk_signal_disconnect_by_data (GTK_OBJECT (tooltipsdata->widget),
  				 (gpointer) tooltipsdata);
   gtk_object_remove_data (GTK_OBJECT (tooltipsdata->widget), tooltips_data_key);
@@ -153,122 +160,40 @@ gtk_tooltips_destroy (GtkObject *object)
 	}
     }
 
-  if (tooltips->tip_window != NULL)
-    {
-      gtk_widget_destroy (tooltips->tip_window);
-      gtk_widget_unref (tooltips->tip_window);
-      tooltips->tip_window = NULL;
-    }
-
-  if (tooltips->gc != NULL)
-    {
-      gdk_gc_destroy (tooltips->gc);
-      tooltips->gc = NULL;
-    }
+  if (tooltips->tip_window)
+    gtk_widget_destroy (tooltips->tip_window);
 }
 
-static void
-gtk_tooltips_create_window (GtkTooltips *tooltips)
+void
+gtk_tooltips_force_window (GtkTooltips *tooltips)
 {
-  tooltips->tip_window = gtk_window_new (GTK_WINDOW_POPUP);
-  gtk_widget_ref (tooltips->tip_window);
-  gtk_window_set_policy (GTK_WINDOW (tooltips->tip_window), FALSE, FALSE, TRUE);
-  gtk_widget_realize (tooltips->tip_window);
-}
+  g_return_if_fail (GTK_IS_TOOLTIPS (tooltips));
 
-static void
-gtk_tooltips_layout_text (GtkTooltips *tooltips, GtkTooltipsData *data)
-{
-  gchar *row_end, *text, *row_text, *break_pos;
-  gint i, row_width, window_width = 0;
-  size_t len;
-
-  if (tooltips->tip_window == NULL)
-    gtk_tooltips_create_window (tooltips);
-
-  g_list_foreach (data->row, gtk_tooltips_free_string, 0);
-  if (data->row)
-    g_list_free (data->row);
-  data->row = 0;
-  data->font = tooltips->tip_window->style->font;
-  data->width = 0;
-
-  text = data->tip_text;
-  if (!text)
-    return;
-
-  while (*text)
+  if (!tooltips->tip_window)
     {
-      row_end = strchr (text, '\n');
-      if (!row_end)
-       row_end = strchr (text, '\0');
+      tooltips->tip_window = gtk_window_new (GTK_WINDOW_POPUP);
+      gtk_widget_set_app_paintable (tooltips->tip_window, TRUE);
+      gtk_window_set_policy (GTK_WINDOW (tooltips->tip_window), FALSE, FALSE, TRUE);
+      gtk_widget_set_name (tooltips->tip_window, "gtk-tooltips");
+      gtk_container_set_border_width (GTK_CONTAINER (tooltips->tip_window), 4);
 
-      len = row_end - text + 1;
-      row_text = g_new(gchar, len);
-      memcpy (row_text, text, len - 1);
-      row_text[len - 1] = '\0';
+      gtk_signal_connect_object (GTK_OBJECT (tooltips->tip_window), 
+				 "expose_event",
+				 GTK_SIGNAL_FUNC (gtk_tooltips_paint_window), 
+				 GTK_OBJECT (tooltips));
 
-      /* now either adjust the window's width or shorten the row until
-        it fits in the window */
+      tooltips->tip_label = gtk_label_new (NULL);
+      gtk_label_set_line_wrap (GTK_LABEL (tooltips->tip_label), TRUE);
+      gtk_misc_set_alignment (GTK_MISC (tooltips->tip_label), 0.5, 0.5);
+      gtk_widget_show (tooltips->tip_label);
+      
+      gtk_container_add (GTK_CONTAINER (tooltips->tip_window), tooltips->tip_label);
 
-      while (1)
-       {
-         row_width = gdk_string_width (data->font, row_text);
-         if (!window_width)
-           {
-             /* make an initial guess at window's width: */
-
-             if (row_width > gdk_screen_width () / 4)
-               window_width = gdk_screen_width () / 4;
-             else
-               window_width = row_width;
-           }
-         if (row_width <= window_width)
-           break;
-
-         if (strchr (row_text, ' '))
-           {
-             /* the row is currently too wide, but we have blanks in
-                the row so we can break it into smaller pieces */
-
-             gint avg_width = row_width / strlen (row_text);
-
-             i = window_width;
-             if (avg_width)
-               i /= avg_width;
-             if ((size_t) i >= len)
-               i = len - 1;
-
-             break_pos = strchr (row_text + i, ' ');
-             if (!break_pos)
-               {
-                 break_pos = row_text + i;
-                 while (*--break_pos != ' ');
-               }
-             *break_pos = '\0';
-           }
-         else
-           {
-             /* we can't break this row into any smaller pieces, so
-                we have no choice but to widen the window: */
-
-             window_width = row_width;
-             break;
-           }
-       }
-      if (row_width > data->width)
-       data->width = row_width;
-      data->row = g_list_append (data->row, row_text);
-      text += strlen (row_text);
-      if (!*text)
-       break;
-
-      if (text[0] == '\n' && text[1])
-       /* end of paragraph and there is more text to come */
-       data->row = g_list_append (data->row, 0);
-      ++text;  /* skip blank or newline */
+      gtk_signal_connect (GTK_OBJECT (tooltips->tip_window),
+			  "destroy",
+			  GTK_SIGNAL_FUNC (gtk_widget_destroyed),
+			  &tooltips->tip_window);
     }
-  data->width += 8;    /* leave some border */
 }
 
 void
@@ -314,79 +239,94 @@ gtk_tooltips_set_tip (GtkTooltips *tooltips,
 {
   GtkTooltipsData *tooltipsdata;
 
-  g_return_if_fail (tooltips != NULL);
   g_return_if_fail (GTK_IS_TOOLTIPS (tooltips));
   g_return_if_fail (widget != NULL);
 
   tooltipsdata = gtk_tooltips_data_get (widget);
-  if (tooltipsdata)
-    gtk_tooltips_widget_remove (tooltipsdata->widget, tooltipsdata);
 
   if (!tip_text)
-    return;
-
-  tooltipsdata = g_new0 (GtkTooltipsData, 1);
-
-  if (tooltipsdata != NULL)
     {
+      if (tooltipsdata)
+	gtk_tooltips_widget_remove (tooltipsdata->widget, tooltipsdata);
+      return;
+    }
+  
+  if (tooltips->active_tips_data 
+      && tooltips->active_tips_data->widget == widget
+      && GTK_WIDGET_DRAWABLE (tooltips->active_tips_data->widget))
+    {
+      g_free (tooltipsdata->tip_text);
+      g_free (tooltipsdata->tip_private);
+
+      tooltipsdata->tip_text = g_strdup (tip_text);
+      tooltipsdata->tip_private = g_strdup (tip_private);
+      
+      gtk_tooltips_draw_tips (tooltips);
+    }
+  else 
+    {
+      gtk_widget_ref (widget);
+      
+      if (tooltipsdata)
+        gtk_tooltips_widget_remove (tooltipsdata->widget, tooltipsdata);
+      
+      tooltipsdata = g_new0 (GtkTooltipsData, 1);
+      
       tooltipsdata->tooltips = tooltips;
       tooltipsdata->widget = widget;
-      gtk_widget_ref (widget);
 
       tooltipsdata->tip_text = g_strdup (tip_text);
       tooltipsdata->tip_private = g_strdup (tip_private);
 
-      gtk_tooltips_layout_text (tooltips, tooltipsdata);
       tooltips->tips_data_list = g_list_append (tooltips->tips_data_list,
-                                             tooltipsdata);
-      gtk_signal_connect_after(GTK_OBJECT (widget), "event",
+                                                tooltipsdata);
+      gtk_signal_connect_after (GTK_OBJECT (widget), "event",
                                (GtkSignalFunc) gtk_tooltips_event_handler,
- 	                       (gpointer) tooltipsdata);
+                               tooltipsdata);
 
       gtk_object_set_data (GTK_OBJECT (widget), tooltips_data_key,
-                           (gpointer) tooltipsdata);
+                           tooltipsdata);
 
       gtk_signal_connect (GTK_OBJECT (widget), "unmap",
                           (GtkSignalFunc) gtk_tooltips_widget_unmap,
-                          (gpointer) tooltipsdata);
+                          tooltipsdata);
 
       gtk_signal_connect (GTK_OBJECT (widget), "unrealize",
                           (GtkSignalFunc) gtk_tooltips_widget_unmap,
-                          (gpointer) tooltipsdata);
+                          tooltipsdata);
 
       gtk_signal_connect (GTK_OBJECT (widget), "destroy",
-			  (GtkSignalFunc) gtk_tooltips_widget_remove,
-			  (gpointer) tooltipsdata);
+                          (GtkSignalFunc) gtk_tooltips_widget_remove,
+                          tooltipsdata);
     }
 }
 
-void
-gtk_tooltips_set_colors (GtkTooltips *tooltips,
-			 GdkColor    *background,
-			 GdkColor    *foreground)
+static gint
+gtk_tooltips_paint_window (GtkTooltips *tooltips)
 {
-  g_return_if_fail (tooltips != NULL);
+  gtk_paint_flat_box (tooltips->tip_window->style, tooltips->tip_window->window,
+		      GTK_STATE_NORMAL, GTK_SHADOW_OUT, 
+		      NULL, GTK_WIDGET(tooltips->tip_window), "tooltip",
+		      0, 0, -1, -1);
 
-  if (background != NULL)
-    tooltips->foreground = foreground;
-  if (foreground != NULL)
-    tooltips->background = background;
+  return FALSE;
 }
 
 static void
 gtk_tooltips_draw_tips (GtkTooltips * tooltips)
 {
+  GtkRequisition requisition;
   GtkWidget *widget;
   GtkStyle *style;
-  gint gap, x, y, w, h, scr_w, scr_h, baseline_skip;
+  gint x, y, w, h, scr_w, scr_h;
   GtkTooltipsData *data;
-  GList *el;
 
   if (!tooltips->tip_window)
-    gtk_tooltips_create_window (tooltips);
+    gtk_tooltips_force_window (tooltips);
   else if (GTK_WIDGET_VISIBLE (tooltips->tip_window))
-    gtk_widget_hide (tooltips->tip_window);
+    g_get_current_time (&tooltips->last_popdown);
 
+  gtk_widget_ensure_style (tooltips->tip_window);
   style = tooltips->tip_window->style;
   
   widget = tooltips->active_tips_data->widget;
@@ -395,27 +335,19 @@ gtk_tooltips_draw_tips (GtkTooltips * tooltips)
   scr_h = gdk_screen_height ();
 
   data = tooltips->active_tips_data;
-  if (data->font != style->font)
-    gtk_tooltips_layout_text (tooltips, data);
 
-  gap = (style->font->ascent + style->font->descent) / 4;
-  if (gap < 2)
-    gap = 2;
-  baseline_skip = style->font->ascent + style->font->descent + gap;
-  w = data->width;
-  h = 8 - gap;
-  for (el = data->row; el; el = el->next)
-    if (el->data)
-      h += baseline_skip;
-    else
-      h += baseline_skip / 2;
-  if (h < 8)
-    h = 8;
+  gtk_label_set_text (GTK_LABEL (tooltips->tip_label), data->tip_text);
+
+  gtk_widget_size_request (tooltips->tip_window, &requisition);
+  w = requisition.width;
+  h = requisition.height;
 
   gdk_window_get_pointer (NULL, &x, NULL, NULL);
   gdk_window_get_origin (widget->window, NULL, &y);
+  if (GTK_WIDGET_NO_WINDOW (widget))
+    y += widget->allocation.y;
 
-  x -= ((w >> 1) + 4);
+  x -= (w / 2 + 4);
 
   if ((x + w) > scr_w)
     x -= (x + w) - scr_w;
@@ -427,51 +359,8 @@ gtk_tooltips_draw_tips (GtkTooltips * tooltips)
   else
     y = y + widget->allocation.height + 4;
 
-  gtk_widget_set_usize (tooltips->tip_window, w + 1, h + 1);
-  gtk_widget_popup (tooltips->tip_window, x, y);
-
-  if (tooltips->gc == NULL)
-    tooltips->gc = gdk_gc_new (tooltips->tip_window->window);
-
-  if (tooltips->background != NULL)
-    {
-      gdk_gc_set_foreground (tooltips->gc, tooltips->background);
-      gdk_gc_set_background (tooltips->gc, tooltips->foreground);
-    }
-  else
-    {
-      gdk_gc_set_foreground (tooltips->gc, &style->bg[GTK_STATE_NORMAL]);
-      gdk_gc_set_background (tooltips->gc, &style->fg[GTK_STATE_NORMAL]);
-    }
-
-  gdk_gc_set_font (tooltips->gc, style->font);
-  gdk_draw_rectangle (tooltips->tip_window->window, tooltips->gc, TRUE, 0, 0, w, h);
-
-  if (tooltips->foreground != NULL)
-    {
-      gdk_gc_set_foreground (tooltips->gc, tooltips->foreground);
-      gdk_gc_set_background (tooltips->gc, tooltips->background);
-    }
-  else
-    {
-      gdk_gc_set_foreground (tooltips->gc, &style->fg[GTK_STATE_NORMAL]);
-      gdk_gc_set_background (tooltips->gc, &style->bg[GTK_STATE_NORMAL]);
-    }
-
-  gdk_draw_rectangle (tooltips->tip_window->window, tooltips->gc, FALSE, 0, 0, w, h);
-  y = style->font->ascent + 4;
-
-  for (el = data->row; el; el = el->next)
-    {
-      if (el->data)
-       {
-         gdk_draw_string (tooltips->tip_window->window, style->font,
-			  tooltips->gc, 4, y, el->data);
-         y += baseline_skip;
-       }
-      else
-       y += baseline_skip / 2;
-    }
+  gtk_window_move (GTK_WINDOW (tooltips->tip_window), x, y);
+  gtk_widget_show (tooltips->tip_window);
 }
 
 static gint
@@ -479,10 +368,14 @@ gtk_tooltips_timeout (gpointer data)
 {
   GtkTooltips *tooltips = (GtkTooltips *) data;
 
+  GDK_THREADS_ENTER ();
+  
   if (tooltips->active_tips_data != NULL &&
       GTK_WIDGET_DRAWABLE (tooltips->active_tips_data->widget))
     gtk_tooltips_draw_tips (tooltips);
 
+  GDK_THREADS_LEAVE ();
+
   return FALSE;
 }
 
@@ -491,7 +384,11 @@ gtk_tooltips_set_active_widget (GtkTooltips *tooltips,
                                 GtkWidget   *widget)
 {
   if (tooltips->tip_window)
-    gtk_widget_hide (tooltips->tip_window);
+    {
+      if (GTK_WIDGET_VISIBLE (tooltips->tip_window))
+	g_get_current_time (&tooltips->last_popdown);
+      gtk_widget_hide (tooltips->tip_window);
+    }
   if (tooltips->timer_tag)
     {
       gtk_timeout_remove (tooltips->timer_tag);
@@ -518,6 +415,22 @@ gtk_tooltips_set_active_widget (GtkTooltips *tooltips,
 	    }
 	}
     }
+  else
+    {
+      tooltips->use_sticky_delay = FALSE;
+    }
+}
+
+static gboolean
+gtk_tooltips_recently_shown (GtkTooltips *tooltips)
+{
+  GTimeVal now;
+  glong msec;
+  
+  g_get_current_time (&now);
+  msec = (now.tv_sec  - tooltips->last_popdown.tv_sec) * 1000 +
+	  (now.tv_usec - tooltips->last_popdown.tv_usec) / 1000;
+  return (msec < STICKY_REVERT_DELAY);
 }
 
 static gint
@@ -551,16 +464,35 @@ gtk_tooltips_event_handler (GtkWidget *widget,
       if (tooltips->enabled &&
 	  (!old_tips_data || old_tips_data->widget != widget))
 	{
+	  guint delay;
+	  
 	  gtk_tooltips_set_active_widget (tooltips, widget);
 	  
-	  tooltips->timer_tag = gtk_timeout_add (tooltips->delay,
+	  if (tooltips->use_sticky_delay  &&
+	      gtk_tooltips_recently_shown (tooltips))
+	    delay = STICKY_DELAY;
+	  else
+	    delay = tooltips->delay;
+	  tooltips->timer_tag = gtk_timeout_add (delay,
 						 gtk_tooltips_timeout,
 						 (gpointer) tooltips);
 	}
       break;
 
+    case GDK_LEAVE_NOTIFY:
+      {
+	gboolean use_sticky_delay;
+
+	use_sticky_delay = tooltips->tip_window &&
+		GTK_WIDGET_VISIBLE (tooltips->tip_window);
+	gtk_tooltips_set_active_widget (tooltips, NULL);
+	tooltips->use_sticky_delay = use_sticky_delay;
+      }
+      break;
+      
     default:
       gtk_tooltips_set_active_widget (tooltips, NULL);
+      return FALSE;
       break;
     }
 
@@ -591,3 +523,25 @@ gtk_tooltips_widget_remove (GtkWidget *widget,
 					    tooltipsdata);
   gtk_tooltips_destroy_data (tooltipsdata);
 }
+
+void
+_gtk_tooltips_show_tip (GtkWidget *widget)
+{
+  /* Showing the tip from the keyboard */
+
+  /* FIXME this function is completely broken right now,
+   * popdown doesn't occur when it should.
+   */
+  
+  GtkTooltipsData *tooltipsdata;
+
+  tooltipsdata = gtk_tooltips_data_get (widget);
+
+  if (tooltipsdata == NULL)
+    return;
+
+  gtk_tooltips_set_active_widget (tooltipsdata->tooltips,
+                                  widget);
+
+  gtk_tooltips_timeout (tooltipsdata->tooltips);
+}
